@@ -1,52 +1,22 @@
 """
 Loading data from audio files.
 
-```
-data, samplingrate = load_audio('audio/file.wav')
-```
-Loads the whole file by trying different modules until it succeeds to load the data.
+- `load_audio()` loads a whole audio file at once.
+- `AudioLoader` allow to read data from audio files in chunks.
+- `unwrap()` unwraps clipped data that are folded into the available data range.
 
-```
-data = AudioLoader('audio/file.wav', 60.0)
-```
-or
-```
-with open_audio_loader('audio/file.wav', 60.0) as data:
-```
-Create an `AudioLoader` instance that loads chuncks of 60 seconds long data on demand.
-data can be used like a read-only numpy array of floats.
+The read in data are always numpy arrays of floats ranging between -1 and 1.
+The arrays are 2-D arrays with first axis time and second axis channel,
+even for single channel data.
 
-In some wav files clipped data are folded into the available data range. With
-```
-data = unwrap(data)
-```
-such data can be unwrapped.
+If an audio file cannot be loaded, you might need to install additional packages.
+See [installation](https://bendalab.github.io/audioio/installation)
+for further instructions.
 
 For a demo run the module as:
 ```
 python -m audioio.audioloader audiofile.wav
 ```
-
-## Installation notes
-
-The built in wave module only supports a few wav file formats.
-The libsndfile library can handle many more formats and is interfaced
-by many different python packages (e.g. SoundFile, wavefile, scikits.audiolab).
-We recommend using SoundFile for best results:
-
-To install SoundFile on Linux do:
-```
-sudo apt-get install -y libsndfile1 libsndfile1-dev libffi-dev
-sudo pip install SoundFile
-```
-
-mp3 and similar formats are supported by the audioread module. Install it via:
-```
-sudo apt-get install -y libav-tools python-audioread
-```
-
-See te audioio.audiomodules module and script for further information on
-installed audio packages and instructions on how to install them.
 """
  
 import warnings
@@ -54,13 +24,14 @@ import os.path
 import numpy as np
 from .audiomodules import *
 
+
 def load_wave(filepath, verbose=0):
     """
     Load wav file using the wave module from pythons standard libray.
     
     Documentation
     -------------
-    https://docs.python.org/2/library/wave.html
+    https://docs.python.org/3.8/library/wave.html
 
     Parameters
     ----------
@@ -79,8 +50,10 @@ def load_wave(filepath, verbose=0):
 
     Raises
     ------
-    ImportError: The wave module is not installed
-    *: Loading of the data failed
+    ImportError
+        The wave module is not installed
+    *
+        Loading of the data failed
     """
     if not audio_modules['wave']:
         raise ImportError
@@ -116,15 +89,6 @@ def load_ewave(filepath, verbose=0):
     Documentation
     -------------
     https://github.com/melizalab/py-ewave
-        
-    Installation
-    ------------
-    ```
-    git clone https://github.com/melizalab/py-ewave
-    cd py-ewave
-    python setup.py build
-    sudo python setup.py install
-    ```
 
     Parameters
     ----------
@@ -142,8 +106,10 @@ def load_ewave(filepath, verbose=0):
 
     Raises
     ------
-    ImportError: The ewave module is not installed
-    *: Loading of the data failed
+    ImportError
+        The ewave module is not installed
+    *
+        Loading of the data failed
     """
     if not audio_modules['ewave']:
         raise ImportError
@@ -184,8 +150,10 @@ def load_wavfile(filepath, verbose=0):
 
     Raises
     ------
-    ImportError: The scipy.io module is not installed
-    *: Loading of the data failed
+    ImportError
+        The scipy.io module is not installed
+    *
+        Loading of the data failed
     """
     if not audio_modules['scipy.io.wavfile']:
         raise ImportError
@@ -214,13 +182,6 @@ def load_soundfile(filepath, verbose=0):
     -------------
     http://pysoundfile.readthedocs.org
     http://www.mega-nerd.com/libsndfile
-        
-    Installation
-    ------------
-    ```
-    sudo apt-get install -y libsndfile1 libsndfile1-dev libffi-dev
-    sudo pip install SoundFile
-    ```
 
     Parameters
     ----------
@@ -238,8 +199,10 @@ def load_soundfile(filepath, verbose=0):
 
     Raises
     ------
-    ImportError: The soundfile module is not installed.
-    *: Loading of the data failed.
+    ImportError
+        The soundfile module is not installed.
+    *
+        Loading of the data failed.
     """
     if not audio_modules['soundfile']:
         raise ImportError
@@ -260,13 +223,6 @@ def load_wavefile(filepath, verbose=0):
     -------------
     https://github.com/vokimon/python-wavefile
 
-    Installation
-    ------------
-    ```
-    sudo apt-get install -y libsndfile1
-    sudo pip install wavefile
-    ```
-
     Parameters
     ----------
     filepath: string
@@ -283,8 +239,10 @@ def load_wavefile(filepath, verbose=0):
 
     Raises
     ------
-    ImportError: The wavefile module is not installed.
-    *: Loading of the data failed.
+    ImportError
+        The wavefile module is not installed.
+    *
+        Loading of the data failed.
     """
     if not audio_modules['wavefile']:
         raise ImportError
@@ -293,65 +251,13 @@ def load_wavefile(filepath, verbose=0):
     return data.astype(np.float64, copy=False).T, float(rate)
 
 
-def load_audiolab(filepath, verbose=0):
-    """
-    Load audio file using scikits.audiolab (based on libsndfile).
-
-    Documentation
-    -------------
-    http://cournape.github.io/audiolab/
-    https://github.com/cournape/audiolab
-                
-    Installation
-    ------------
-    ```
-    sudo apt-get install -y libsndfile1
-    sudo pip install scikits.audiolab
-    ```
-
-    Parameters
-    ----------
-    filepath: string
-        The full path and name of the file to load.
-    verbose: int
-        If >0 show detailed error/warning messages (not used).
-
-    Returns
-    -------
-    data: array
-        All data traces as an 2-D numpy array, first dimension is time, second is channel.
-    rate: float
-        The sampling rate of the data in Hertz.
-
-    Raises
-    ------
-    ImportError: The scikits.audiolab module is not installed.
-    *: Loading of the data failed.
-    """
-    if not audio_modules['scikits.audiolab']:
-        raise ImportError
-
-    af = audiolab.Sndfile(filepath, 'r')
-    rate = af.samplerate
-    data = af.read_frames(af.nframes)
-    if len(data.shape) == 1:
-        data = np.reshape(data,(-1, 1))
-    return data, float(rate)
-
-
 def load_audioread(filepath, verbose=0):
     """
     Load audio file using audioread.
 
     Documentation
     -------------
-    https://github.com/sampsyo/audioread
-
-    Installation
-    ------------
-    sudo apt-get install -y libav-tools python-audioread
-        
-    audioread is not available in python x,y.
+    https://github.com/beetbox/audioread
 
     Parameters
     ----------
@@ -369,8 +275,10 @@ def load_audioread(filepath, verbose=0):
 
     Raises
     ------
-    ImportError: The audioread module is not installed.
-    *: Loading of the data failed.
+    ImportError
+        The audioread module is not installed.
+    *
+        Loading of the data failed.
     """
     if not audio_modules['audioread']:
         raise ImportError
@@ -395,20 +303,24 @@ def load_audioread(filepath, verbose=0):
     return data/(2.0**15-1.0), float(rate)
 
 
-# list of implemented load functions:
-audio_loader = [
-    ['soundfile', load_soundfile],
-    ['audioread', load_audioread],
-    ['wave', load_wave],
-    ['wavefile', load_wavefile],
-    ['scikits.audiolab', load_audiolab],
-    ['ewave', load_ewave],
-    ['scipy.io.wavfile', load_wavfile]
-    ]
+audio_loader_funcs = (
+    ('soundfile', load_soundfile),
+    ('audioread', load_audioread),
+    ('wave', load_wave),
+    ('wavefile', load_wavefile),
+    ('ewave', load_ewave),
+    ('scipy.io.wavfile', load_wavfile)
+    )
+""" List of implemented load functions.
+
+Each element of the list is a tuple with the module's name and the load function.
+"""    
+
 
 def load_audio(filepath, verbose=0):
     """
     Call this function to load all channels of audio data from a file.
+    
     This function tries different python modules to load the audio file.
 
     Parameters
@@ -421,33 +333,49 @@ def load_audio(filepath, verbose=0):
     Returns
     -------
     data: array
-        All data traces as an 2-D numpy array, first dimension is time, second is channel.
+        All data traces as an 2-D numpy array, even for single channel data.
+        First dimension is time, second is channel.
+        Data values range maximally between -1 and 1.
     rate: float
         The sampling rate of the data in Hertz.
 
     Raises
     ------
-    ValueError: Empty `filepath`.
-    FileNotFoundError: `filepath` is not an existing file.
-    EOFError: File size of `filepath` is zero.
-    IOError: Failed to load data.
+    ValueError
+        Empty `filepath`.
+    FileNotFoundError
+        `filepath` is not an existing file.
+    EOFError
+        File size of `filepath` is zero.
+    IOError
+        Failed to load data.
+
+    Example
+    -------
+    ```
+    import matplotlib.pyplot as plt
+    from audioio import load_audio
+    
+    data, rate = load_audio('some/audio.wav')
+    plt.plot(np.arange(len(data))/rate, data[:,0])
+    plt.show()
+    ```
     """
     # check values:
-    if len(filepath) == 0:
+    if filepath is None or len(filepath) == 0:
         raise ValueError('input argument filepath is empty string!')
     if not os.path.isfile(filepath):
-        FileNotFoundException = getattr(__builtins__,'FileNotFoundError', IOError)
-        raise FileNotFoundException('file "%s" not found' % filepath)
+        raise FileNotFoundError('file "%s" not found' % filepath)
     if os.path.getsize(filepath) <= 0:
         raise EOFError('file "%s" is empty (size=0)!' % filepath)
 
     # load an audio file by trying various modules:
     success = False
     not_installed = []
-    for lib, load_file in audio_loader:
+    for lib, load_file in audio_loader_funcs:
         if not audio_modules[lib]:
             if verbose > 1:
-                print('failed to load data from file "%s" using %s module' %
+                print('unable to load data from file "%s" using %s module: module not available' %
                       (filepath, lib))
             not_installed.append(lib)
             continue
@@ -460,6 +388,7 @@ def load_audio(filepath, verbose=0):
                           (filepath, lib))
                     if verbose > 1:
                         print('  sampling rate: %g Hz' % rate)
+                        print('  channels     : %d' % data.shape[1])
                         print('  data values  : %d' % len(data))
                 break
         except Exception as e:
@@ -488,7 +417,7 @@ def unwrap(data):
 
     Returns
     -------
-    data: as input data
+    data: same as input data
         The fixed data.
     """
     if len(data.shape) > 1:
@@ -507,7 +436,8 @@ def unwrap(data):
 
 class AudioLoader(object):
     """Buffered reading of audio data for random access of the data in the file.
-    This allows for reading very large audio files that do not fit into memory.
+    
+    The class allows for reading very large audio files that do not fit into memory.
     An AudioLoader instance can be used like a huge read-only numpy array, i.e.
     ```
     data = AudioLoader('path/to/audio/file.wav')
@@ -531,7 +461,7 @@ class AudioLoader(object):
     -----
     ```
     import audioio as aio
-    with aio.open_audio_loader(filepath, 60.0, 10.0) as data:
+    with aio.AudioLoader(filepath, 60.0, 10.0) as data:
         # do something with the content of the file:
         x = data[0:10000]
         y = data[10000:20000]
@@ -570,9 +500,19 @@ class AudioLoader(object):
     ```
     for supported and available modules.
     
+    Parameters
+    ----------
+    filepath: string
+        Name of the file.
+    buffersize: float
+        Size of internal buffer in seconds.
+    backsize: float
+        Part of the buffer to be loaded before the requested start index in seconds.
+    verbose: int
+        If >0 show detailed error/warning messages.
 
-    Member variables
-    ----------------
+    Attributes
+    ----------
     samplerate: float
         The sampling rate of the data in seconds.
     channels: int
@@ -582,37 +522,34 @@ class AudioLoader(object):
     shape: tuple
         Frames and channels of the data.
 
-    Some member functions
-    ---------------------
-    len(): the number of frames
-    open(): open an audio file by trying available audio modules.
-    open_*(): open an audio file with the respective audio module.
-    close(): close the file.
+    Methods
+    -------
+    len()
+        Number of frames.
+    open()
+        Open an audio file by trying available audio modules.
+    open_*()
+        Open an audio file with the respective audio module.
+    close()
+        Close the file.
 
-    TODO
-    ----
-    Access via __getitem__ or __next__ is slow!
+    Notes
+    -----
+    Access via `__getitem__` or `__next__` is slow!
     Even worse, using numpy functions on this class first converts
     it to a numpy array - that is something we actually do not want!
     We should subclass directly from numpy.ndarray .
     For details see http://docs.scipy.org/doc/numpy/user/basics.subclassing.html
     When subclassing, there is an offset argument, that might help to
-    speed up __getitem__ .
+    speed up `__getitem__` .
     """
     
     def __init__(self, filepath=None, buffersize=10.0, backsize=0.0, verbose=0):
-        """Initialize the AudioLoader instance. If filepath is not None open the file.
+        """Initialize the AudioLoader instance.
 
-        Parameters
-        ----------
-        filepath: string
-            Name of the file.
-        buffersize: float
-            Size of internal buffer in seconds.
-        backsize: float
-            Part of the buffer to be loaded before the requested start index in seconds.
-        verbose: int
-            If >0 show detailed error/warning messages.
+        If filepath is not None open the file.
+
+        See documentation of the class for more information.
         """
         self.sf = None
         self.samplerate = 0.0
@@ -793,7 +730,8 @@ class AudioLoader(object):
 
         Raises
         ------
-        ImportError: The wave module is not installed
+        ImportError
+            The wave module is not installed
         """
         self.verbose = verbose
         if self.verbose > 1:
@@ -877,7 +815,8 @@ class AudioLoader(object):
 
         Raises
         ------
-        ImportError: The ewave module is not installed.
+        ImportError
+            The ewave module is not installed.
         """
         self.verbose = verbose
         if self.verbose > 1:
@@ -946,7 +885,8 @@ class AudioLoader(object):
 
         Raises
         ------
-        ImportError: The SoundFile module is not installed
+        ImportError
+            The SoundFile module is not installed
         """
         self.verbose = verbose
         if self.verbose > 1:
@@ -1015,7 +955,8 @@ class AudioLoader(object):
 
         Raises
         ------
-        ImportError: The wavefile module is not installed
+        ImportError
+            The wavefile module is not installed
         """
         self.verbose = verbose
         if self.verbose > 1:
@@ -1065,74 +1006,6 @@ class AudioLoader(object):
                       % (self.buffer.shape[0], self.offset, self.offset+self.buffer.shape[0]))
 
             
-    # scikits.audiolab interface:        
-    def open_audiolab(self, filepath, buffersize=10.0, backsize=0.0, verbose=0):
-        """Open audio file for reading using the scikits.audiolab module.
-
-        Parameters
-        ----------
-        filepath: string
-            Name of the file.
-        buffersize: float
-            Size of internal buffer in seconds.
-        backsize: float
-            Part of the buffer to be loaded before the requested start index in seconds.
-        verbose: int
-            If >0 show detailed error/warning messages.
-
-        Raises
-        ------
-        ImportError: The scikits.audiolab module is not installed
-        """
-        self.verbose = verbose
-        if self.verbose > 1:
-            print('open_audiolab(filepath) with filepath=%s' % filepath)
-        if not audio_modules['scikits.audiolab']:
-            self.samplerate = 0.0
-            self.channels = 0
-            self.frames = 0
-            self.shape = (0, 0)
-            self.offset = 0
-            raise ImportError
-        if self.sf is not None:
-            self._close_audiolab()
-        self.sf = audiolab.Sndfile(filepath, 'r')
-        self.samplerate = float(self.sf.samplerate)
-        self.channels = self.sf.channels
-        self.frames = int(self.sf.nframes)
-        self.shape = (self.frames, self.channels)
-        self.buffersize = int(buffersize*self.samplerate)
-        self.backsize = int(backsize*self.samplerate)
-        self._init_buffer()
-        self.offset = 0
-        self.close = self._close_audiolab
-        self._update_buffer = self._update_buffer_audiolab
-        return self
-
-    def _close_audiolab(self):
-        """ Close the audio file using the scikits.audiolab module. """
-        if self.sf is not None:
-            self.sf.close()
-            self.sf = None
-
-    def _update_buffer_audiolab(self, start, stop):
-        """Make sure that the buffer contains the data between
-        start and stop using the scikits.audiolab module.
-        """
-        if start < self.offset or stop > self.offset + self.buffer.shape[0]:
-            offset, size = self._read_indices(start, stop)
-            r_offset, r_size = self._recycle_buffer(offset, size)
-            self.sf.seek(r_offset)  # undocumented ...
-            buffer = self.sf.read_frames(nframes=r_size, dtype=self.buffer.dtype)
-            if len(buffer.shape) == 1:
-                buffer = np.reshape(buffer,(-1, 1))
-            self.buffer[r_offset-offset:r_offset+r_size-offset,:] = buffer
-            self.offset = offset
-            if self.verbose > 0:
-                print('  loaded %d frames from %d up to %d'
-                      % (self.buffer.shape[0], self.offset, self.offset+self.buffer.shape[0]))
-                
-            
     # audioread interface:        
     def open_audioread(self, filepath, buffersize=10.0, backsize=0.0, verbose=0):
         """Open audio file for reading using the audioread module.
@@ -1153,7 +1026,8 @@ class AudioLoader(object):
 
         Raises
         ------
-        ImportError: The audioread module is not installed
+        ImportError
+            The audioread module is not installed
         """
         self.verbose = verbose
         if self.verbose > 1:
@@ -1297,10 +1171,14 @@ class AudioLoader(object):
 
         Raises
         ------
-        ValueError: Empty `filepath`.
-        FileNotFoundError: `filepath` is not an existing file.
-        EOFError: File size of `filepath` is zero.
-        IOError: Failed to load data.
+        ValueError
+            Empty `filepath`.
+        FileNotFoundError
+            `filepath` is not an existing file.
+        EOFError
+            File size of `filepath` is zero.
+        IOError
+            Failed to load data.
         """
         self.buffer = np.array([])
         self.samplerate = 0.0
@@ -1317,7 +1195,6 @@ class AudioLoader(object):
             ['audioread', self.open_audioread],
             ['wave', self.open_wave],
             ['wavefile', self.open_wavefile],
-            ['scikits.audiolab', self.open_audiolab],
             ['ewave', self.open_ewave]
             ]
         # open an audio file by trying various modules:
@@ -1351,66 +1228,39 @@ class AudioLoader(object):
                           (filepath, need_install))
         return self
 
-
-open_audio_loader = AudioLoader
-                
-
-if __name__ == "__main__":
-    import sys
-    import matplotlib.pyplot as plt
-
-    print("Checking audioloader module ...")
-    print('')
-    print('Usage:')
-    print('  python audioloader.py [-p] <audio/file.wav>')
-    print('  -p: plot data')
-
-    plot = False
-    if len(sys.argv) > 1 and sys.argv[1] == '-p':
-        plot = True
-
-    #disable_module('soundfile')
-    #disable_module('audioread')
-                
-    filepath = sys.argv[-1]
     
+def demo(file_path, plot):
+    """ Demo of the audioloader functions.
+
+    Parameters
+    ----------
+    file_path: string
+        File path of an audio file.
+    plot: bool
+        If True also plot the loaded data.
+    """
     print('')
     print("try load_audio:")
-    full_data, rate = load_audio(filepath, 2)
+    full_data, rate = load_audio(file_path, 1)
     if plot:
         plt.plot(np.arange(len(full_data))/rate, full_data[:,0])
         plt.show()
 
-    print('')
-    for lib, load_file in audio_loader:
-        if not audio_modules[lib]:
-            continue
-        try:
-            data, rate = load_file(filepath, 1)
-            print('loaded data from file "%s" with %s' %
-                    (filepath, lib))
-        except Exception as e:
-            print('failed to load data from file "%s" with %s' %
-                  (filepath, lib))
-    
-    print('')
-    print("cross check:")
-    data1, rate1 = load_soundfile(filepath)
-    data2, rate2 = load_audioread(filepath)
-    n = min((len(data1), len(data2)))
-    print("rms difference is %g" % np.std(data1[:n]-data2[:n]))
-    if plot:
-        plt.plot(np.arange(len(data1))/rate1, data1[:,0])
-        plt.plot(np.arange(len(data2))/rate2, data2[:,0])
-        plt.show()
+    if audio_modules['soundfile'] and audio_modules['audioread']:
+        print('')
+        print("cross check:")
+        data1, rate1 = load_soundfile(file_path)
+        data2, rate2 = load_audioread(file_path)
+        n = min((len(data1), len(data2)))
+        print("rms difference is %g" % np.std(data1[:n]-data2[:n]))
+        if plot:
+            plt.plot(np.arange(len(data1))/rate1, data1[:,0])
+            plt.plot(np.arange(len(data2))/rate2, data2[:,0])
+            plt.show()
     
     print('')
     print("try AudioLoader:")
-    #data = AudioLoader()
-    #data.open_audioread(filepath, 2.0, 1.0, 2)
-    #data = AudioLoader(filepath, 8.0, 3.0, 2)
-    #with data.open_soundfile(filepath, 2.0, 1.0, 2):
-    with open_audio_loader(filepath, 4.0, 1.0, 0) as data:
+    with AudioLoader(file_path, 4.0, 1.0, 1) as data:
         print('samplerate: %g' % data.samplerate)
         print('channels: %d %d' % (data.channels, data.shape[1]))
         print('frames: %d %d' % (len(data), data.shape[0]))
@@ -1446,4 +1296,53 @@ if __name__ == "__main__":
             if plot:
                 plt.plot((i+np.arange(len(x)))/rate, x)
                 plt.show()
-    #data.close()
+
+
+def main(args):
+    """ Call demo with command line arguments.
+
+    Parameters
+    ----------
+    args: list of strings
+        Command line arguments as provided by sys.argv
+    """
+    print("Checking audioloader module ...")
+
+    help = False
+    plot = False
+    file_path = None
+    mod = False
+    for arg in args[1:]:
+        if mod:
+            if not select_module(arg):
+                print('can not select module %s that is not installed' % arg)
+                return
+            mod = False
+        elif arg == '-h':
+            help = True
+            break
+        elif arg == '-p':
+            plot = True
+        elif arg == '-m':
+            mod = True
+        else:
+            file_path = arg
+            break
+
+    if help:
+        print('')
+        print('Usage:')
+        print('  python -m audioio.audioloader [-m <module>] [-p] <audio/file.wav>')
+        print('  -m: audio module to be used')
+        print('  -p: plot loaded data')
+        return
+
+    if plot:
+        import matplotlib.pyplot as plt
+
+    demo(file_path, plot)
+
+
+if __name__ == "__main__":
+    import sys
+    main(sys.argv)
