@@ -1,24 +1,16 @@
 """
 Writing numpy arrays of floats to audio files.
 
-```
-write_audio('audio/file.wav', data, samplerate)
-```
-Writes the whole file at once with an installed audio module that
-supports the requested file format.
+- `write_audio()`: write audio data to file.
+- `available_formats()`: audio file formats supported by any of the installed audio modules.
+- `available_encodings()`: encodings of an audio file format supported by any of the installed audio modules.
 
-```
-available_formats()
-available_encodings()
-```
-return lists of supported formats and encodings.
+The data to be written are 1-D or 2-D numpy arrays of floats ranging between -1 and 1
+with first axis time and second axis channel.
 
-We recommend SoundFile for best results:
-Installation:
-```
-sudo apt-get install libsndfile1 libsndfile1-dev libffi-dev
-sudo pip install SoundFile
-```
+For support of more audio formats, you might need to install additional packages.
+See [installation](https://bendalab.github.io/audioio/installation)
+for further instructions.
 
 For a demo, run the script as:
 ```
@@ -43,6 +35,7 @@ def formats_wave():
     else:
         return ['WAV']
 
+
 def encodings_wave(format):
     """Encodings of an audio file format supported by the wave module.
 
@@ -62,20 +55,21 @@ def encodings_wave(format):
         return []
     else:
         return ['PCM_32', 'PCM_16', 'PCM_U8']
-    
+
+
 def write_wave(filepath, data, samplerate, format=None, encoding=None):
     """
     Write audio data using the wave module from pythons standard libray.
     
     Documentation
     -------------
-    https://docs.python.org/2/library/wave.html
+    https://docs.python.org/3.8/library/wave.html
 
     Parameters
     ----------
     filepath: string
         Full path and name of the file to write.
-    data: 1d- or 2d-array of floats
+    data: 1-D or 2-D array of floats
         Array with the data (first index time, second index channel,
         values within -1.0 and 1.0).
     samplerate: float
@@ -83,20 +77,23 @@ def write_wave(filepath, data, samplerate, format=None, encoding=None):
     format: string or None
         File format, only 'WAV' is supported.
     encoding: string or None
-        Encoding of the data: 'PCM_32', PCM_16', or 'PCM_U8'
+        Encoding of the data: 'PCM_32', 'PCM_16', or 'PCM_U8'
+        If None or empty string use 'PCM_16'.
 
     Raises
     ------
-    ImportError: The wave module is not installed.
-    *: Writing of the data failed.
-    ValueError: File format or encoding not supported.
+    ImportError
+        The wave module is not installed.
+    *
+        Writing of the data failed.
+    ValueError
+        File format or encoding not supported.
     """
     if not audio_modules['wave']:
         raise ImportError
 
-    if format is not None and format.upper() != 'WAV':
-        raise ValueError('file format %s not supported by wave-module' % format)
-        return
+    if format and format.upper() != 'WAV':
+        raise ValueError('file format %s not supported by wave module' % format)
 
     wave_encodings = {'PCM_32': [4, 'i4'],
                       'PCM_16': [2, 'i2'],
@@ -107,8 +104,7 @@ def write_wave(filepath, data, samplerate, format=None, encoding=None):
         encoding = 'PCM_16'
     encoding = encoding.upper()
     if not encoding in wave_encodings:
-        raise ValueError('file encoding %s not supported by wave-module' % encoding)
-        return
+        raise ValueError('file encoding %s not supported by wave module' % encoding)
     sampwidth = wave_encodings[encoding][0]
     dtype = wave_encodings[encoding][1]
 
@@ -127,7 +123,10 @@ def write_wave(filepath, data, samplerate, format=None, encoding=None):
     else:
         buffer = np.floor(data * factor).astype(dtype)
         buffer[data >= 1.0] = factor - 1
-    wf.writeframes(buffer.tostring())
+    try:
+        wf.writeframes(buffer.tobytes())
+    except AttributeError: 
+        wf.writeframes(buffer.tostring())
     wf.close()
 
 
@@ -144,6 +143,7 @@ def formats_ewave():
     else:
         return ['WAV', 'WAVEX']
 
+
 def encodings_ewave(format):
     """Encodings of an audio file format supported by the ewave module.
 
@@ -163,7 +163,8 @@ def encodings_ewave(format):
         return []
     else:
         return ['PCM_64', 'PCM_32', 'PCM_16', 'FLOAT', 'DOUBLE']
-    
+
+
 def write_ewave(filepath, data, samplerate, format=None, encoding=None):
     """
     Write audio data using the ewave module from pythons standard libray.
@@ -176,28 +177,31 @@ def write_ewave(filepath, data, samplerate, format=None, encoding=None):
     ----------
     filepath: string
         Full path and name of the file to write.
-    data: 1d- or 2d-array of floats
+    data: 1-D or 2-D array of floats
         Array with the data (first index time, second index channel,
         values within -1.0 and 1.0).
     samplerate: float
         Sampling rate of the data in Hertz.
     format: string or None
-        File format, only 'WAV' is supported.
+        File format, only 'WAV' and 'WAVEX' are supported.
     encoding: string or None
         Encoding of the data: 'PCM_64', 'PCM_32', PCM_16', 'FLOAT', 'DOUBLE'
+        If None or empty string use 'PCM_16'.
 
     Raises
     ------
-    ImportError: The ewave module is not installed.
-    *: Writing of the data failed.
-    ValueError: File format or encoding not supported.
+    ImportError
+        The ewave module is not installed.
+    *
+        Writing of the data failed.
+    ValueError
+        File format or encoding not supported.
     """
     if not audio_modules['ewave']:
         raise ImportError
 
-    if format is not None and format.upper() != 'WAV' and format.upper() != 'WAVEX':
-        raise ValueError('file format %s not supported by ewave-module' % format)
-        return
+    if format and format.upper() != 'WAV' and format.upper() != 'WAVEX':
+        raise ValueError('file format %s not supported by ewave module' % format)
 
     ewave_encodings = {'PCM_64': 'l',
                        'PCM_32': 'i',
@@ -210,8 +214,7 @@ def write_ewave(filepath, data, samplerate, format=None, encoding=None):
         encoding = 'PCM_16'
     encoding = encoding.upper()
     if not encoding in ewave_encodings:
-        raise ValueError('file encoding %s not supported by ewave-module' % encoding)
-        return
+        raise ValueError('file encoding %s not supported by ewave module' % encoding)
 
     channels = 1
     if len(data.shape) > 1:
@@ -235,6 +238,7 @@ def formats_wavfile():
     else:
         return ['WAV']
 
+
 def encodings_wavfile(format):
     """Encodings of an audio file format supported by the scipy.io.wavfile module.
 
@@ -254,7 +258,8 @@ def encodings_wavfile(format):
         return []
     else:
         return ['PCM_U8', 'PCM_16', 'PCM_32', 'PCM_64', 'FLOAT', 'DOUBLE']
-    
+
+
 def write_wavfile(filepath, data, samplerate, format=None, encoding=None):
     """
     Write audio data using the scipy.io.wavfile module.
@@ -267,7 +272,7 @@ def write_wavfile(filepath, data, samplerate, format=None, encoding=None):
     ----------
     filepath: string
         Full path and name of the file to write.
-    data: 1d- or 2d-array of floats
+    data: 1-D or 2-D array of floats
         Array with the data (first index time, second index channel,
         values within -1.0 and 1.0).
     samplerate: float
@@ -275,20 +280,23 @@ def write_wavfile(filepath, data, samplerate, format=None, encoding=None):
     format: string or None
         File format, only 'WAV' is supported.
     encoding: string or None
-        Encoding of the data: PCM_16
+        Encoding of the data: 'PCM_64', 'PCM_32', PCM_16', 'PCM_U8', 'FLOAT', 'DOUBLE'
+        If None or empty string use 'PCM_16'.
 
     Raises
     ------
-    ImportError: The wavfile module is not installed.
-    *: Writing of the data failed.
-    ValueError: File format or encoding not supported.
+    ImportError
+        The wavfile module is not installed.
+    ValueError
+        File format or encoding not supported.
+    *
+        Writing of the data failed.
     """
     if not audio_modules['scipy.io.wavfile']:
         raise ImportError
 
-    if format is not None and format.upper() != 'WAV':
-        raise ValueError('file format %s not supported by scipy.io.wavfile-module' % format)
-        return
+    if format and format.upper() != 'WAV':
+        raise ValueError('file format %s not supported by scipy.io.wavfile module' % format)
 
     wave_encodings = {'PCM_U8': [1, 'u1'],
                       'PCM_16': [2, 'i2'],
@@ -302,8 +310,7 @@ def write_wavfile(filepath, data, samplerate, format=None, encoding=None):
         encoding = 'PCM_16'
     encoding = encoding.upper()
     if not encoding in wave_encodings:
-        raise ValueError('file encoding %s not supported by scipy.io.wavfile-module' % encoding)
-        return
+        raise ValueError('file encoding %s not supported by scipy.io.wavfile module' % encoding)
     sampwidth = wave_encodings[encoding][0]
     dtype = wave_encodings[encoding][1]
     if sampwidth == 1:
@@ -331,7 +338,8 @@ def formats_soundfile():
         return []
     else:
         return sorted(list(soundfile.available_formats()))
-            
+
+
 def encodings_soundfile(format):
     """Encodings of an audio file format supported by the SoundFile module.
 
@@ -349,7 +357,8 @@ def encodings_soundfile(format):
         return []
     else:
         return sorted(list(soundfile.available_subtypes(format)))
-        
+
+
 def write_soundfile(filepath, data, samplerate, format=None, encoding=None):
     """
     Write audio data using the SoundFile module (based on libsndfile).
@@ -362,7 +371,7 @@ def write_soundfile(filepath, data, samplerate, format=None, encoding=None):
     ----------
     filepath: string
         Full path and name of the file to write.
-    data: 1d- or 2d-array of floats
+    data: 1-D or 2-D array of floats
         Array with the data (first index time, second index channel,
         values within -1.0 and 1.0).
     samplerate: float
@@ -371,23 +380,29 @@ def write_soundfile(filepath, data, samplerate, format=None, encoding=None):
         File format.
     encoding: string or None
         Encoding of the data.
+        If None or empty string use 'PCM_16'.
 
     Raises
     ------
-    ImportError: The SoundFile module is not installed.
-    *: Writing of the data failed.
+    ImportError
+        The SoundFile module is not installed.
+    *
+        Writing of the data failed.
     """
     if not audio_modules['soundfile']:
         raise ImportError
 
     if format is not None:
-        format = format.upper()
+        if len(format) == 0:
+            format = None
+        else:
+            format = format.upper()
 
     if encoding == '':
         encoding = None
-    if encoding is not None:
-        encoding = encoding.upper()
-
+    if encoding is None:
+        encoding = 'PCM_16'
+    encoding = encoding.upper()
     soundfile.write(filepath, data, int(samplerate), format=format, subtype=encoding)
 
 
@@ -410,6 +425,7 @@ def formats_wavefile():
             formats.append(attr)
     return sorted(formats)
 
+
 def encodings_wavefile(format):
     """Encodings supported by the wavefile module.
 
@@ -435,6 +451,7 @@ def encodings_wavefile(format):
              and v != wavefile.Format.SUBMASK ):
             encodings.append(attr)
     return sorted(encodings)
+
     
 def write_wavefile(filepath, data, samplerate, format=None, encoding=None):
     """
@@ -448,7 +465,7 @@ def write_wavefile(filepath, data, samplerate, format=None, encoding=None):
     ----------
     filepath: string
         Full path and name of the file to write.
-    data: 1d- or 2d-array of floats
+    data: 1-D or 2-D array of floats
         Array with the data (first index time, second index channel,
         values within -1.0 and 1.0).
     samplerate: float
@@ -457,11 +474,16 @@ def write_wavefile(filepath, data, samplerate, format=None, encoding=None):
         File format as in wavefile.Format.
     encoding: string or None
         Encoding of the data as in wavefile.Format.
+        If None or empty string use 'PCM_16'.
 
     Raises
     ------
-    ImportError: The wavefile module is not installed.
-    *: Writing of the data failed.
+    ImportError
+        The wavefile module is not installed.
+    ValueError
+        File format or encoding not supported.
+    *
+        Writing of the data failed.
     """
     if not audio_modules['wavefile']:
         raise ImportError
@@ -472,128 +494,46 @@ def write_wavefile(filepath, data, samplerate, format=None, encoding=None):
         format = filepath.split('.')[-1]
         # TODO: we need a mapping from file extensions to formats and default encoding!
     format = format.upper()
-    format_value = getattr(wavefile.Format, format)
+    try:
+        format_value = getattr(wavefile.Format, format)
+    except AttributeError:
+        raise ValueError('file format %s not supported by wavefile module' % format)
 
     if encoding is None:
         encoding = ''
     if len(encoding) == 0:
         encodings = encodings_wavefile(format)
-        # TODO: better default settings!
         if 'PCM_16' in encodings:
             encoding = 'PCM_16'
         else:
             encoding = encodings[0]
     encoding = encoding.upper()
-    encoding_value = getattr(wavefile.Format, encoding)
+    try:
+        encoding_value = getattr(wavefile.Format, encoding)
+    except AttributeError:
+        raise ValueError('file encoding %s not supported by wavefile module' % encoding)
         
     channels = 1
     if len(data.shape) > 1:
         channels = data.shape[1]
+    else:
+        data = data.reshape((-1, 1))
     with wavefile.WaveWriter(filepath, channels=channels, samplerate=int(samplerate),
                              format=format_value|encoding_value) as w:
         w.write(data.T)
 
 
-def formats_audiolab():
-    """Audio file formats supported by the scikits.audiolab module.
+audio_formats_funcs = (
+    ('soundfile', formats_soundfile),
+    ('wavefile', formats_wavefile),
+    ('wave', formats_wave),
+    ('ewave', formats_ewave),
+    ('scipy.io.wavfile', formats_wavfile)
+    )
+""" List of implemented formats functions.
 
-    Returns
-    -------
-    formats: list of strings
-        List of supported file formats as strings.
-    """
-    if not audio_modules['scikits.audiolab']:
-        return []
-    formats = [f.upper() for f in audiolab.available_file_formats()]
-    return sorted(formats)
-
-def encodings_audiolab(format):
-    """Encodings of an audio file format supported by the scikits.audiolab module.
-
-    Parameters
-    ----------
-    format: str
-        The file format.
-
-    Returns
-    -------
-    encodings: list of strings
-        List of supported encodings as strings.
-    """
-    if not audio_modules['scikits.audiolab']:
-        return []
-    try:
-        encodings = []
-        for encoding in audiolab.available_encodings(format.lower()):
-            if encoding[0:3] == 'pcm':
-                encoding = 'pcm_' + encoding[3:]
-            if encoding == 'float32': encoding = 'float'
-            if encoding == 'float64': encoding = 'double'
-            encodings.append(encoding.upper())
-        return sorted(encodings)
-    except ValueError:
-        return []
-
-def write_audiolab(filepath, data, samplerate, format=None, encoding=None):
-    """
-    Write audio data using the scikits.audiolab module (based on libsndfile).
-
-    Documentation
-    -------------
-    http://cournape.github.io/audiolab/
-    https://github.com/cournape/audiolab
-
-    Parameters
-    ----------
-    filepath: string
-        Full path and name of the file to write.
-    data: 1d- or 2d-array of floats
-        Array with the data (first index time, second index channel,
-        values within -1.0 and 1.0).
-    samplerate: float
-        Sampling rate of the data in Hertz.
-    format: string or None
-        File format like the SF_FORMAT_ constants of libsndfile.
-    encoding: string or None
-        Encoding of the data like the SF_FORMAT_ constants of libsndfile.
-
-    Raises
-    ------
-    ImportError: The scikits.audiolab module is not installed.
-    *: Writing of the data failed.
-    """
-    if not audio_modules['scikits.audiolab']:
-        raise ImportError
-
-    if format is None:
-        format = ''
-    if len(format) == 0:
-        format = filepath.split('.')[-1].lower()
-        # TODO: we need a mapping from file extensions to formats and default encoding!
-    format = format.lower()
-
-    if encoding is None:
-        encoding = ''
-    if len(encoding) == 0:
-        encodings = encodings_audiolab(format)
-        # TODO: better default settings!
-        if 'PCM_16' in encodings:
-            encoding = 'PCM_16'
-        else:
-            encoding = encodings[0]
-    encoding = encoding.lower()
-    encoding = encoding.replace('pcm_', 'pcm')
-    if encoding == 'float': encoding = 'float32'
-    if encoding == 'double': encoding = 'float64'
-        
-    channels = 1
-    if len(data.shape) > 1:
-        channels = data.shape[1]
-
-    af = audiolab.Sndfile(filepath, 'w', format=audiolab.Format(format, encoding),
-                          channels=channels, samplerate=int(samplerate))
-    af.write_frames(data)
-    af.close()
+Each element of the list is a tuple with the module's name and the formats function.
+"""
 
 
 def available_formats():
@@ -604,13 +544,25 @@ def available_formats():
     formats: list of strings
         List of supported file formats as strings.
     """
-    audio_formats = [formats_wave, formats_ewave, formats_wavfile,
-                     formats_soundfile, formats_wavefile, formats_audiolab]
     formats = set()
-    for formats_func in audio_formats:
+    for module, formats_func in audio_formats_funcs:
         formats |= set(formats_func())
     return sorted(list(formats))
 
+
+audio_encodings_funcs = (
+    ('soundfile', encodings_soundfile),
+    ('wavefile', encodings_wavefile),
+    ('wave', encodings_wave),
+    ('ewave', encodings_ewave),
+    ('scipy.io.wavfile', encodings_wavfile)
+    )
+""" List of implemented encodings functions.
+
+Each element of the list is a tuple with the module's name and the encodings function.
+"""
+
+
 def available_encodings(format):
     """Encodings of an audio file format supported by any of the installed audio modules.
 
@@ -624,22 +576,32 @@ def available_encodings(format):
     encodings: list of strings
         List of supported encodings as strings.
     """
-    audio_encodings = [encodings_wave, encodings_ewave, encodings_wavfile,
-                        encodings_soundfile, encodings_audiolab, encodings_wavefile]
-    first_sndfilelib_inx = 3
-    wavefile_inx = 5
     got_sndfile = False
     encodings = set()
-    for e_inx, encodings_func in enumerate(audio_encodings):
-        if e_inx == wavefile_inx and got_sndfile:
+    for module, encodings_func in audio_encodings_funcs:
+        if got_sndfile and module == 'scipy.io.wavfile':
             continue
         encs = encodings_func(format)
-        if e_inx >= first_sndfilelib_inx and len(encs) > 0:
-            got_sndfile = True
         encodings |= set(encs)
+        if module in ['soundfile', 'wavefile'] and len(encs) > 0:
+            got_sndfile = True
     return sorted(list(encodings))
 
-def write_audio(filepath, data, samplerate, format=None, encoding=None):
+
+audio_writer_funcs = (
+    ('soundfile', write_soundfile),
+    ('wavefile', write_wavefile),
+    ('wave', write_wave),
+    ('ewave', write_ewave),
+    ('scipy.io.wavfile', write_wavfile)
+    )
+""" List of implemented write functions.
+
+Each element of the list is a tuple with the module's name and the write function.
+"""
+
+
+def write_audio(filepath, data, samplerate, format=None, encoding=None, verbose=0):
     """
     Write audio data to file.
 
@@ -647,42 +609,59 @@ def write_audio(filepath, data, samplerate, format=None, encoding=None):
     ----------
     filepath: string
         Full path and name of the file to write.
-    data: 1d- or 2d-array of floats
+    data: 1-D or 2-D array of floats
         Array with the data (first index time, second index channel,
         values within -1.0 and 1.0).
     samplerate: float
         Sampling rate of the data in Hertz.
     format: string or None
-        File format.
+        File format. If None deduce file format from filepath.
+        See `available_formats()` for possible values.
     encoding: string or None
-        Encoding of the data.
+        Encoding of the data. See `available_encodings()` for possible values.
+        If None or empty string use 'PCM_16'.
+    verbose: int
+        If >0 show detailed error/warning messages.
 
     Raises
     ------
-    ValueError: `filepath` is empty string.
-    IOError: Writing of the data failed.
-    """
+    ValueError
+        `filepath` is empty string.
+    IOError
+        Writing of the data failed.
 
-    audio_writer = [
-        ['soundfile', write_soundfile],
-        ['scikits.audiolab', write_audiolab],
-        ['wavefile', write_wavefile],
-        ['wave', write_wave],
-        ['ewave', write_ewave],
-        ['scipy.io.wavfile', write_wavfile]
-        ]
+    Example
+    -------
+    ```
+    import numpy as np
+    from audioio import write_audio
+    
+    samplerate = 28000.0
+    freq = 800.0
+    time = np.arange(0.0, 1.0, 1/samplerate) # one second
+    data = np.sin(2.0*np.p*freq*time)        # 800Hz sine wave
+    write_audio('audio/file.wav', data, samplerate)
+    ```
+    """
 
     if len(filepath) == 0:
         raise ValueError('input argument filepath is empty string!')
 
     # write audio file by trying available modules:
     success = False
-    for lib, write_file in audio_writer:
+    for lib, write_file in audio_writer_funcs:
         if not audio_modules[lib]:
             continue
         try:
             write_file(filepath, data, samplerate, format, encoding)
             success = True
+            if verbose > 0:
+                print('wrote data to file "%s" using %s module' %
+                      (filepath, lib))
+                if verbose > 1:
+                    print('  sampling rate: %g Hz' % samplerate)
+                    print('  channels     : %d' % (data.shape[1] if len(data.shape) > 1 else 1))
+                    print('  frames       : %d' % len(data))
             break
         except Exception as e:
             pass
@@ -690,30 +669,68 @@ def write_audio(filepath, data, samplerate, format=None, encoding=None):
         raise IOError('failed to write data to file "%s"' % filepath)
 
 
-if __name__ == "__main__":
-    import sys
-    import numpy as np
+def demo(file_path, encoding=''):
+    """ Demo of the audiowriter functions.
 
-    print("Checking audiowriter module ...")
+    Parameters
+    ----------
+    file_path: string
+        File path of an audio file.
+    encoding: string
+        Encoding to be used.
+    """
     print('')
-    print('Usage:')
-    print('  python audiowriter.py <filename> <encoding>')
-
-    filepath = 'test.wav'
-    if len(sys.argv) > 1:
-        filepath = sys.argv[1]
-    encoding = ''
-    if len(sys.argv) > 2:
-        encoding = sys.argv[2]
-
+    print('generate data ...')
     samplerate = 44100.0
-    t = np.arange(int(2*samplerate))/samplerate
+    t = np.arange(0.0, 2.0, 1.0/samplerate)
     data = np.sin(2.0*np.pi*880.0*t)
         
-    print('')
-    print("write_audio(%s) ..." % filepath)
-    write_audio(filepath, data, samplerate, encoding=encoding)
+    print("write_audio(%s) ..." % file_path)
+    write_audio(file_path, data, samplerate, encoding=encoding, verbose=2)
 
     print('done.')
     
 
+def main(args):
+    """ Call demo with command line arguments.
+
+    Parameters
+    ----------
+    args: list of strings
+        Command line arguments as provided by sys.argv
+    """
+    print("Checking audiowriter module ...")
+
+    help = False
+    file_path = None
+    encoding = ''
+    mod = False
+    for arg in args[1:]:
+        if mod:
+            select_module(arg)
+            mod = False
+        elif arg == '-h':
+            help = True
+            break
+        elif arg == '-m':
+            mod = True
+        elif file_path is None:
+            file_path = arg
+        else:
+            encoding = arg
+            break
+    if file_path is None:
+        file_path = 'test.wav'
+
+    if help:
+        print('')
+        print('Usage:')
+        print('  python -m audioio.audiowriter [-m module] [<filename>] [<encoding>]')
+        return
+
+    demo(file_path, encoding)
+
+
+if __name__ == "__main__":
+    import sys
+    main(sys.argv)
