@@ -1,18 +1,25 @@
 """
 Play numpy arrays as audio.
 
-The globally defined functions
-```
-play(data, rate)
-beep(duration, frequeny)
-```
-use a global instance of the `PlayAudio` class to play a sound
-on the default audio output.
+- `play()`: playback audio data.
+- `beep()`: playback a tone.
+- `PlayAudio()`: audio playback.
+- `fade_in()`: fade in a signal in place.
+- `fade_out()`: fade out a signal in place.
+- `fade()`: fade in an out a signal in place.
+- `note2freq()`: convert textual note to corresponding frequency.
+
+Accepted data for playback are 1-D or 2-D numpy arrays with values ranging from -1 to 1.
+If necessary data are downsampled automatically to match supported sampling rates.
+
+The globally defined functions `play()` and `beep()`
+use the global instance `handle` of the `PlayAudio` class to play a sound
+on the default audio output device.
 
 Alternatively you may use the `PlayAudio` class directly, like this:
 ```
-with open_audio_player() as audio:
-    audio.beep(1.0, 440.0)
+with PlayAudio() as audio:
+    audio.beep()
 ```
 or without context management:
 ```
@@ -21,51 +28,34 @@ audio.beep(1.0, 'a4')
 audio.close()
 ```
 
-The `note2freq()` function converts a musical note, like 'f#4',
-to the corresponding frequency.
-The `beep()` functions also accept notes for the frequency argument,
-and uses `note2freq()` to get the right frequency.
-
-Data can be multiplied with a squared-sine for fading in and out with
-`fade_in()`, `fade_out()`, and `fade()`.
+You might need to install additional packages for better audio output.
+See [installation](https://bendalab.github.io/audioio/installation)
+for further instructions.
 
 For a demo, run the script as:
 ```
 python -m audioio.playaudio
 ```
-
-See also:
-https://wiki.python.org/moin/Audio/
-https://docs.python.org/3/library/mm.html
-
-The modules supports the standard modules `ossaudiodev` for Linux and `winsound` for Windows.
-However, we recommend to install the `portaudio` library and the `pyaudio` module for
-better performance.
-
-On Linux do:
-```
-sudo apt-get install -y libportaudio2 portaudio19-dev python-pyaudio python3-pyaudio
-```
 """
 
+from sys import platform
 import os
 import warnings
-import time
 import numpy as np
-try:
-    from cStringIO import StringIO
-except ImportError:
-    from io import StringIO
+from time import sleep
+from io import BytesIO
 from multiprocessing import Process
 from .audiomodules import *
 
 
-# default audio device handler:
 handle = None
+""" Default audio device handler.
+
+Will get an PlayAudio instance assigned. """
 
 
 def note2freq(note, a4freq=440.0):
-    """Converts textual note to the corresponding frequency.
+    """Convert textual note to corresponding frequency.
 
     Parameters
     ----------
@@ -78,7 +68,7 @@ def note2freq(note, a4freq=440.0):
         The last character specifies the octave.
         'a4' is defined by a4freq.
     a4freq: float
-        The frequency of a4.
+        The frequency of a4 in Hertz.
 
     Returns
     -------
@@ -95,19 +85,23 @@ def note2freq(note, a4freq=440.0):
     octave = 4
     if not isinstance(note, str) or len(note) == 0:
         raise ValueError('no note specified')
+    # note:
     if note[0] < 'a' or note[0] > 'g':
         raise ValueError('invalid note %s' % note[0])
-    # note:
     index = 0
     tonemap = [0, 2, 3, 5, 7, 8, 10]
     tone = tonemap[ord(note[index]) - ord('a')]
     index += 1
     # flat or sharp:
+    flat  = False
+    sharp = False
     if index < len(note):
         if note[index] == 'b':
+            flat = True
             tone -= 1
             index += 1
         elif note[index] == '#':
+            sharp = True
             tone += 1
             index += 1
     # octave:
@@ -117,7 +111,11 @@ def note2freq(note, a4freq=440.0):
             octave *= 10
             octave += ord(note[index]) - ord('0')
             index += 1
-    if tone >= 3:
+    # remaining characters:
+    if index < len(note):
+        raise ValueError('invalid characters in note %s' % note)
+    # compute frequency:
+    if (tone >= 3 and not sharp) or (tone == 2 and flat):
         octave -= 1
     tone += 12*(octave-4)
     # frequency:
@@ -127,15 +125,15 @@ def note2freq(note, a4freq=440.0):
 
 def fade_in(data, rate, fadetime):
     """
-    Fade the signal in.
+    Fade in a signal in place.
 
-    The first fadetime seconds of the data are multiplied with a squared sine in place.
+    The first `fadetime` seconds of the data are multiplied with a squared sine in place.
     
     Parameters
     ----------
     data: array
         The data to be faded in, either 1-D array for single channel output,
-        or 2-day array with first axis time and second axis channel.
+        or 2-D array with first axis time and second axis channel.
     rate: float
         The sampling rate in Hertz.
     fadetime: float
@@ -152,15 +150,15 @@ def fade_in(data, rate, fadetime):
         
 def fade_out(data, rate, fadetime):
     """
-    Fade the signal out.
+    Fade out a signal in place.
 
-    The last fadetime seconds of the data are multiplied with a squared sine in place.
+    The last `fadetime` seconds of the data are multiplied with a squared sine in place.
     
     Parameters
     ----------
     data: array
         The data to be faded out, either 1-D array for single channel output,
-        or 2-day array with first axis time and second axis channel.
+        or 2-D array with first axis time and second axis channel.
     rate: float
         The sampling rate in Hertz.
     fadetime: float
@@ -177,16 +175,16 @@ def fade_out(data, rate, fadetime):
 
 def fade(data, rate, fadetime):
     """
-    Fade the signal in and out.
+    Fade in and out a signal in place.
 
-    The first and last fadetime seconds of the data are multiplied with
+    The first and last `fadetime` seconds of the data are multiplied with
     a squared sine in place.
         
     Parameters
     ----------
     data: array
         The data to be faded, either 1-D array for single channel output,
-        or 2-day array with first axis time and second axis channel.
+        or 2-D array with first axis time and second axis channel.
     rate: float
         The sampling rate in Hertz.
     fadetime: float
@@ -197,14 +195,46 @@ def fade(data, rate, fadetime):
 
 
 class PlayAudio(object):
+    """ Audio playback.
+
+    Parameters
+    ----------
+    verbose: int
+        Verbosity level.
+
+    Methods
+    -------
+    play(data, rate, scale=None, blocking=True)
+        Playback audio data.
+    beep(duration=0.5, frequency=880.0, amplitude=0.5, rate=44100.0, fadetime=0.05, blocking=True)
+        Playback a pure tone.
+    open()
+        Initialize the PlayAudio class with the best module available.
+    close()
+        Terminate module for playing audio.
+    stop()
+        Stop any playback in progress.
+
+    Examples
+    --------
+    ```
+    from audioio import PlayAudio
+    
+    with PlayAudio() as audio:
+        audio.beep()
+    ```
+    or without context management:
+    ```
+    audio = PlayAudio()
+    audio.beep(1.0, 'a4')
+    audio.close()
+    ```
+    """
     
     def __init__(self, verbose=0):
         """Initialize module for playing audio.
 
-        Parameters
-        ----------
-        verbose: int
-            Verbosity level.
+        See `PlayAudio` documentation for details.
         """
         self.verbose = verbose
         self.handle = None
@@ -214,11 +244,11 @@ class PlayAudio(object):
         self.open()
 
     def _close(self):
-        """Terminate module for playing audio."""
+        """Terminate PlayAudio class for playing audio."""
         pass
 
     def _stop(self):
-        """Stop playing."""
+        """Stop any playback in progress."""
         pass
 
     def _play(self, blocking=True):
@@ -226,18 +256,19 @@ class PlayAudio(object):
         pass
 
     def play(self, data, rate, scale=None, blocking=True):
-        """Play audio data.
+        """Playback audio data.
 
         Parameters
         ----------
         data: array
             The data to be played, either 1-D array for single channel output,
-            or 2-day array with first axis time and second axis channel.
+            or 2-D array with first axis time and second axis channel.
+            Data values range between -1 and 1.
         rate: float
             The sampling rate in Hertz.
         scale: float
             Multiply data with scale before playing.
-            If None scale it to the maximum value, if 1.0 do not scale.
+            If `None` scale it to the maximum value, if 1.0 do not scale.
         blocking: boolean
             If False do not block. 
 
@@ -260,13 +291,13 @@ class PlayAudio(object):
             mind = np.abs(np.min(rawdata))
             scale = 1.0/max((mind, maxd))
         rawdata *= scale
-        self.data = np.floor(rawdata*(2**15-1)).astype(np.int16)
+        self.data = np.floor(rawdata*(2**15-1)).astype(np.int16, order='C')
         self.index = 0
         self._do_play(blocking)
 
-    def beep(self, duration, frequency, amplitude=1.0, rate=44100.0,
+    def beep(self, duration=0.5, frequency=880.0, amplitude=0.5, rate=44100.0,
              fadetime=0.05, blocking=True):
-        """Play a pure tone of a given duration and frequency.
+        """Playback a pure tone.
 
         Parameters
         ----------
@@ -277,7 +308,7 @@ class PlayAudio(object):
             If string, a musical note like 'f#5'.
             See note2freq() for details.
         amplitude: float
-            The ampliude of the tone in the range from 0.0 to 1.0.
+            The ampliude (volume) of the tone in the range from 0.0 to 1.0.
         rate: float
             The sampling rate in Hertz.
         fadetime: float
@@ -317,17 +348,19 @@ class PlayAudio(object):
         else:
             dt0 = 1.0/self.rate
             dt1 = scale/self.rate
-            t1 = (len(self.data)+0.5)*dt0
-            new_time = np.arange(0.0, t1, dt1)
-            old_time = np.arange(0.0, t1, dt0)
+            old_time = np.arange(len(self.data))*dt0
+            new_time = np.arange(0.0, old_time[-1]+0.5*dt0, dt1)
             if len(self.data.shape) > 1:
-                data = np.zeros((len(newtime), channels))
+                data = np.zeros((len(new_time), channels), order='C')
                 for c in range(channels):
                     data[:, c] = np.interp(new_time, old_time, self.data[:, c])
+            else:
+                data = np.interp(new_time, old_time, self.data)
+            if self.data.dtype == data.dtype and flags['C_CONTIGUOUS']:
                 self.data = data
             else:
-                self.data = np.interp(new_time, old_time, self.data)
-        if self.verbose:
+                self.data = np.asarray(data, dtype=self.data.dtype, order='C')
+        if self.verbose > 0:
             print('adapted sampling rate from %g Hz down to %g Hz' %
                   (self.rate, self.rate/scale))
         self.rate /= scale
@@ -355,12 +388,21 @@ class PlayAudio(object):
         Documentation
         -------------
         https://people.csail.mit.edu/hubert/pyaudio/
+        http://www.portaudio.com/
 
         Installation
         ------------
         ```
         sudo apt-get install -y libportaudio2 portaudio19-dev python-pyaudio python3-pyaudio
         ```
+        
+        On Windows, download an appropriate (latest version, 32 or 64 bit) wheel from
+        <https://www.lfd.uci.edu/~gohlke/pythonlibs/#pyaudio>.  Install this file with pip,
+        that is go to the folder where the wheel file is downloaded and run
+        ```
+        pip install PyAudio-0.2.11-cp39-cp39-win_amd64.whl
+        ```
+        replace the wheel file name by the one you downloaded.
         """
         if not audio_modules['pyaudio']:
             raise ImportError
@@ -425,12 +467,13 @@ class PlayAudio(object):
                     self.data[index+nr:] = 0
                 if nr > 0:
                     for k in range(nr) :
-                        self.data[index+(nr-k-1)] *= np.sin(0.5*np.pi*float(k)/float(nr))**2.0
-                time.sleep(2*fadetime)
+                        self.data[index+(nr-k-1)] = np.array(self.data[index+(nr-k-1)] *
+                                np.sin(0.5*np.pi*float(k)/float(nr))**2.0, np.int16, order='C')
+                sleep(2*fadetime)
             if self.stream.is_active():
                 self.run = False
                 while self.stream.is_active():
-                    time.sleep(0.01)
+                    sleep(0.01)
                 self.stream.stop_stream()
             self.stream.close()
             self.stream = None
@@ -490,7 +533,10 @@ class PlayAudio(object):
         self.stream.start_stream()
         if blocking:
             while self.stream.is_active():
-                time.sleep(0.01)
+                try:
+                    sleep(0.01)
+                except ValueError:
+                    pass
             self.run = False
             self.stream.stop_stream()
             self.stream.close()
@@ -588,7 +634,8 @@ class PlayAudio(object):
                     self.data[index+nr:] = 0
                 if nr > 0:
                     for k in range(nr) :
-                        self.data[index+(nr-k-1)] *= np.sin(0.5*np.pi*float(k)/float(nr))**2.0
+                        self.data[index+(nr-k-1)] = np.array(self.data[index+(nr-k-1)] *
+                                np.sin(0.5*np.pi*float(k)/float(nr))**2.0, np.int16, order='C')
                 sounddevice.sleep(int(2000*fadetime))
             if self.stream.active:
                 self.run = False
@@ -635,7 +682,7 @@ class PlayAudio(object):
                 success = True
                 break
             except sounddevice.PortAudioError as pae:
-                if pae[1] != -9997:
+                if pae.args[1] != -9997:
                     raise
                 elif self.verbose > 0:
                     print('invalid sampling rate of %g Hz' % rate)
@@ -668,7 +715,74 @@ class PlayAudio(object):
         self._do_play = self._play
         self.stop = self._stop
 
+        
+    def open_simpleaudio(self):
+        """Initialize audio output via simpleaudio package.
 
+        Raises
+        ------
+        ImportError: simpleaudio module is not available.
+
+        Documentation
+        -------------
+        https://simpleaudio.readthedocs.io
+        """
+        if not audio_modules['simpleaudio']:
+            raise ImportError
+        self.handle = True
+        self._do_play = self._play_simpleaudio
+        self.close = self._close_simpleaudio
+        self.stop = self._stop_simpleaudio
+        return self
+
+    def _stop_simpleaudio(self):
+        """Stop any ongoing activity of the simpleaudio package."""
+        if self.handle is not None and self.handle is not True:
+            self.handle.stop()
+    
+    def _play_simpleaudio(self, blocking=True):
+        """
+        Play audio data using the simpleaudio package.
+
+        Parameters
+        ----------
+        blocking: boolean
+            If False do not block. 
+
+        Raises
+        ------
+        ValueError: Invalid sampling rate (after some attemps of resampling).
+        """
+        rates = [self.rate, 44100, 22050]
+        scales = [1, None, None]
+        success = False
+        for rate, scale in zip(rates, scales):
+            if scale is None:
+                scale = self.rate/float(rate)
+            if scale != 1:
+                self._down_sample(self.channels, scale)
+            try:
+                self.handle = simpleaudio.play_buffer(self.data, self.channels,
+                                                      2, int(self.rate))
+                success = True
+                break
+            except ValueError:
+                if self.verbose > 0:
+                    print('invalid sampling rate of %g Hz' % rate)
+        if not success:
+            raise ValueError('No valid sampling rate found')
+        elif blocking:
+            self.handle.wait_done()
+        
+    def _close_simpleaudio(self):
+        """Close audio output using simpleaudio package. """
+        self._stop_simpleaudio()
+        simpleaudio.stop_all()
+        self.handle = None
+        self._do_play = self._play
+        self.stop = self._stop
+
+                
     def open_ossaudiodev(self):
         """Initialize audio output via ossaudiodev module.
 
@@ -719,7 +833,7 @@ class PlayAudio(object):
         """Play the data using the ossaudiodev module."""
         self.osshandle.writeall(self.data)
         if self.run:
-            time.sleep(0.5)
+            sleep(0.5)
             self.osshandle.close()
             self.osshandle = None
             self.run = False
@@ -768,7 +882,7 @@ class PlayAudio(object):
         if blocking:
             self.run = True
             self.osshandle.writeall(self.data)
-            time.sleep(0.5)
+            sleep(0.5)
             self.osshandle.close()
             self.run = False
             self.osshandle = None
@@ -796,22 +910,22 @@ class PlayAudio(object):
 
         Documentation
         -------------
-        https://docs.python.org/2/library/winsound.html
+        https://docs.python.org/3.6/library/winsound.html
         https://mail.python.org/pipermail/tutor/2012-September/091529.html
         """
-        if not audio_modules['winsound'] or not not audio_modules['wave']:
+        if not audio_modules['winsound'] or not audio_modules['wave']:
             raise ImportError
         self.handle = True
         self._do_play = self._play_winsound
         self.close = self._close_winsound
         self.stop = self._stop_winsound
+        self.audio_file = ''
         return self
 
     def _stop_winsound(self):
         """Stop any ongoing activity of the winsound module."""
         winsound.PlaySound(None, winsound.SND_MEMORY)
     
-    
     def _play_winsound(self, blocking=True):
         """
         Play audio data using the winsound module.
@@ -821,40 +935,61 @@ class PlayAudio(object):
         blocking: boolean
             If False do not block. 
         """
-        # write data as wav file to memory:
-        # TODO: check this code!!!
-        self.data_buffer = StringIO()
-        w = wave.open(self.data_buffer, 'w')
-        w.setnchannels(self.channels)
-        w.setsampwidth(2)
-        w.setframerate(int(self.rate))
-        w.setnframes(len(self.data))
-        w.writeframesraw(self.data.tostring())
-        w.close() # TODO: close here or after PlaySound?
         # play file:
         if blocking:
+            # write data as wav file to memory:
+            self.data_buffer = BytesIO()
+            w = wave.open(self.data_buffer, 'w')
+            w.setnchannels(self.channels)
+            w.setsampwidth(2)
+            w.setframerate(int(self.rate))
+            w.setnframes(len(self.data))
+            try:
+                w.writeframes(self.data.tobytes())
+            except AttributeError:
+                w.writeframes(self.data.tostring())
+            w.close()
             winsound.PlaySound(self.data_buffer.getvalue(), winsound.SND_MEMORY)
         else:
-            winsound.PlaySound(self.data_buffer.getvalue(),
-                               winsound.SND_MEMORY | winsound.SND_ASYNC)
+            if self.verbose > 0:
+                print('Warning: asynchronous playback is limited to playing wav files by the winsound module. Install an alternative package as recommended by the audiomodules script. ')
+            # write data as wav file to file:
+            self.audio_file = 'audioio-async_playback.wav'
+            w = wave.open(self.audio_file, 'w')
+            w.setnchannels(self.channels)
+            w.setsampwidth(2)
+            w.setframerate(int(self.rate))
+            w.setnframes(len(self.data))
+            try:
+                w.writeframes(self.data.tobytes())
+            except AttributeError:
+                w.writeframes(self.data.tostring())
+            w.close()
+            winsound.PlaySound(self.audio_file, winsound.SND_ASYNC)
         
     def _close_winsound(self):
         """Close audio output using winsound module. """
         self._stop_winsound()
         self.handle = None
+        if len(self.audio_file) > 0 and os.path.isfile(self.audio_file):
+            os.remove(self.audio_file)
         self._do_play = self._play
         self.stop = self._stop
 
 
     def open(self):
-        """Initialize the audio module with the best module available."""
+        """Initialize the PyAudio class with the best module available."""
         # list of implemented play functions:
         audio_open = [
-            ['pyaudio', self.open_pyaudio],
             ['sounddevice', self.open_sounddevice],
+            ['pyaudio', self.open_pyaudio],
+            ['simpleaudio', self.open_simpleaudio],
             ['ossaudiodev', self.open_ossaudiodev],
             ['winsound', self.open_winsound]
             ]
+        if platform[0:3] == "win":
+            sa = audio_open.pop(2)
+            audio_open.insert(0, sa)
         # open audio device by trying various modules:
         success = False
         for lib, open_device in audio_open:
@@ -870,17 +1005,14 @@ class PlayAudio(object):
                 break
             except Exception as e:
                 if self.verbose > 0:
-                    print('failed to open %s module for playing' % lib)
+                    print('failed to open %s module for playing:' % lib, str(e))
         if not success:
             warnings.warn('cannot open any device for audio output')
         return self
 
 
-open_audio_player = PlayAudio
-                
-
 def play(data, rate, scale=None, blocking=True, verbose=0):
-    """Play audio data.
+    """Playback audio data.
 
     Create an PlayAudio instance on the global variable handle.
 
@@ -888,12 +1020,13 @@ def play(data, rate, scale=None, blocking=True, verbose=0):
     ----------
     data: array
         The data to be played, either 1-D array for single channel output,
-        or 2-day array with first axis time and second axis channel.
+        or 2-D array with first axis time and second axis channel.
+        Data values range between -1 and 1.
     rate: float
         The sampling rate in Hertz.
     scale: float
         Multiply data with scale before playing.
-        If None scale it to the maximum value, if 1.0 do not scale.
+        If `None` scale it to the maximum value, if 1.0 do not scale.
     blocking: boolean
         If False do not block. 
     verbose: int
@@ -905,10 +1038,10 @@ def play(data, rate, scale=None, blocking=True, verbose=0):
     handle.play(data, rate, scale, blocking)
 
     
-def beep(duration, frequency, amplitude=1.0, rate=44100.0,
+def beep(duration=0.5, frequency=880.0, amplitude=0.5, rate=44100.0,
          fadetime=0.05, blocking=True, verbose=0):
     """
-    Play a tone of a given duration and frequency.
+    Playback a tone.
 
     Create an PlayAudio instance on the globale variable handle.
 
@@ -921,7 +1054,7 @@ def beep(duration, frequency, amplitude=1.0, rate=44100.0,
         If string, a musical note like 'f#5'.
         See note2freq() for details
     amplitude: float
-        The ampliude of the tone from 0.0 to 1.0.
+        The ampliude (volume) of the tone from 0.0 to 1.0.
     rate: float
         The sampling rate in Hertz.
     fadetime: float
@@ -936,24 +1069,26 @@ def beep(duration, frequency, amplitude=1.0, rate=44100.0,
         handle = PlayAudio(verbose)
     handle.beep(duration, frequency, amplitude, rate, fadetime, blocking)
 
-    
-if __name__ == "__main__":
+
+def demo():
+    """ Demonstrate the playaudio module.
+    """
     print('play mono beep 1')
-    audio = PlayAudio()
+    audio = PlayAudio(verbose=2)
     audio.beep(1.0, 440.0)
     audio.close()
     
     print('play mono beep 2')
-    with open_audio_player() as audio:
+    with PlayAudio() as audio:
         audio.beep(1.0, 'b4', 0.75, blocking=False)
         print('  done')
-        time.sleep(0.3)
-    time.sleep(0.5)
+        sleep(0.3)
+    sleep(0.5)
 
     print('play mono beep 3')
     beep(1.0, 'c5', 0.25, blocking=False)
     print('  done')
-    time.sleep(0.5)
+    sleep(0.5)
             
     print('play stereo beep')
     duration = 1.0
@@ -965,14 +1100,40 @@ if __name__ == "__main__":
     fade(data, rate, 0.1)
     play(data, rate)
 
-    exit()
 
-    print('play notes')
-    o = 6
-    for oo in range(4):
-        for t in range(7):
-            if t == 2:
-                o += 1
-            tone = '%s%d' % (chr(ord('a')+t), o)
-            print('%-3s %7.1f Hz' % (tone, note2freq(tone)))
-            beep(0.5, tone)
+def main(args):
+    """ Call demo with command line arguments.
+
+    Parameters
+    ----------
+    args: list of strings
+        Command line arguments as provided by sys.argv
+    """
+    help = False
+    mod = False
+    for arg in args[1:]:
+        if mod:
+            if not select_module(arg):
+                return
+            mod = False
+        elif arg == '-h':
+            help = True
+            break
+        elif arg == '-m':
+            mod = True
+        else:
+            break
+
+    if help:
+        print('')
+        print('Usage:')
+        print('  python -m audioio.playaudio [-m <module>]')
+        print('  -m: audio module to be used')
+        return
+        
+    demo()
+
+            
+if __name__ == "__main__":
+    import sys
+    main(sys.argv)
